from django.shortcuts import get_object_or_404
import requests
from src.ChatTokenBuilder2 import ChatTokenBuilder
from crea_app.models import CustomUser, PrivacySettings
from notifications_app.views import send_push_notification
from notifications_app.models import Notification
from rest_framework import viewsets, status, permissions
from rest_framework.response import Response
from rest_framework.views import APIView
from django.utils import timezone
from .models import Follow, FollowRequest, Block, Report
from .serializers import BlockListSerializer, FollowRequestSerializer, FollowSerializer, BlockSerializer, FollowerFollowingSerializer, ReasonSerializer, ReportSerializer
from django.contrib.auth import get_user_model
from userprofile_app.models import Profile, Role
from rest_framework.permissions import IsAuthenticated
from decouple import config
from django.conf import settings
User = get_user_model()

APP_ID = config('APP_ID')
APP_CERTIFICATE = config('APP_CERTIFICATE')

def is_blocked(user, target_user):
    """ Check if 'target_user' is blocked by 'user'. """
    return Block.objects.filter(blocker=user, blocked=target_user).exists() or Block.objects.filter(blocker=target_user, blocked=user).exists()

class ListFollowersView(APIView):
    permission_classes = [permissions.IsAuthenticated]

    def get(self, request, user_id):
        try:
            user = User.objects.get(id=user_id)
        except User.DoesNotExist:
            return Response({"error": "User not found"}, status=status.HTTP_404_NOT_FOUND)

        followers = Follow.objects.filter(following=user)
        data = []
        for follower in followers:
            follower_user = follower.follower
            try:
                follower_profile = Profile.objects.get(user=follower_user)
                role_name = follower_profile.selected_role.role_name if follower_profile.selected_role else ""
                profile_pic_url = follower_profile.profile_image or ""
            except Profile.DoesNotExist:
                role_name = ""
                profile_pic_url = ""
            data.append({
                "follower_id": str(follower_user.id),
                "name": follower_user.username or "",
                "role": role_name,
                "profile_pic_url": profile_pic_url
            })

        return Response({
            "user_id": str(user_id),
            "followers": data
        }, status=status.HTTP_200_OK)

class ListFollowingView(APIView):
    permission_classes = [permissions.IsAuthenticated]

    def get(self, request, user_id):
        try:
            user = User.objects.get(id=user_id)
        except User.DoesNotExist:
            return Response({"error": "User not found"}, status=status.HTTP_404_NOT_FOUND)

        following = Follow.objects.filter(follower=user)
        data = []
        for follow in following:
            following_user = follow.following
            try:
                following_profile = Profile.objects.get(user=following_user)
                role_name = following_profile.selected_role.role_name if following_profile.selected_role else ""
                profile_pic_url = following_profile.profile_image or ""

            except Profile.DoesNotExist:
                role_name = ""
                profile_pic_url = ""
            data.append({
                "followed_id": str(following_user.id),
                "name": following_user.username or "",
                "role": role_name,
                "profile_pic_url": profile_pic_url
            })

        return Response({
            "user_id": str(user_id),
            "following": data
        }, status=status.HTTP_200_OK)

class FollowRequestViewSet(viewsets.ViewSet):
    permission_classes = [permissions.IsAuthenticated]

    def send_request(self, request, user_id=None):
        requester = request.user
        try:
            recipient = User.objects.get(id=user_id)
        except User.DoesNotExist:
            return Response({"error": "User not found"}, status=status.HTTP_404_NOT_FOUND)

        if is_blocked(requester, recipient):
            return Response({"error": "You cannot follow this user."}, status=status.HTTP_403_FORBIDDEN)

        # Prevent self-follow requests
        if int(requester.id) == int(user_id):
            return Response({"error": "You cannot send a follow request to yourself."}, status=status.HTTP_400_BAD_REQUEST)

        # Check if the recipient exists
        try:
            recipient = User.objects.get(id=user_id)
        except User.DoesNotExist:
            return Response({"error": "User not found"}, status=status.HTTP_404_NOT_FOUND)

        # Check if requester is blocked by the recipient
        if Block.objects.filter(blocker=recipient, blocked=requester).exists():
            return Response({"error": "You are blocked by this user."}, status=status.HTTP_403_FORBIDDEN)

        # Check if there's an existing follow request or follow relationship
        follow_request = FollowRequest.objects.filter(requester=requester, recipient=recipient).first()
        follow_exists = Follow.objects.filter(follower=requester, following=recipient).exists()

        if follow_exists:
            return Response({"error": "User is already following this user"}, status=status.HTTP_400_BAD_REQUEST)

        if follow_request:
            # If the request is pending, don't allow another request to be sent
            if follow_request.status == 'pending':
                return Response({"error": "Follow request is already pending"}, status=status.HTTP_400_BAD_REQUEST)

            # If the request was previously accepted or canceled, update it to pending
            follow_request.status = 'pending'
            follow_request.save()
            Notification.objects.filter(user=recipient, sender=requester, action_type="follow_request").delete()
            Notification.objects.create(
                user=recipient,
                sender=requester,
                action_type="follow_request",
                message=f"{request.user.full_name} re-sent you a follow request."
            )
            response = send_push_notification(
                fcm_token=recipient.fcm_token,
                title="New Follow Request",
                body=f"{requester.full_name} sent you a follow request."
            )
            print(response)
            return Response({
                "message": "Follow request re-sent successfully",
                "push_notification": response,
                "data": {
                    "user_id": str(recipient.id),
                    "follower_id": str(requester.id),
                    "requested_at": follow_request.created_at.isoformat()
                }
            }, status=status.HTTP_200_OK)

        # Create a new follow request if none exists
        follow_request = FollowRequest.objects.create(requester=requester, recipient=recipient)
        Notification.objects.filter(user=recipient, sender=requester, action_type="follow_request").delete()
        Notification.objects.create(
            user=recipient,
            sender=requester,
            action_type="follow_request",
            message=f"{request.user.full_name} sent you a follow request."
        )
        print(recipient.fcm_token)
        privacy_settings, created = PrivacySettings.objects.get_or_create(user=recipient)
        if privacy_settings.enable_notifications:
        # if recipient.privacy_settings.enable_notifications:
            response = send_push_notification(
                    fcm_token=recipient.fcm_token,
                    title="New Follow Request",
                    body=f"{requester.full_name} sent you a follow request."
                )
            print(response)
        else:
            response = None 

        return Response({
            "message": "Follow request sent successfully",
            "push_notification": response,
            "data": {
                "user_id": str(recipient.id),
                "follower_id": str(requester.id),
                "requested_at": follow_request.created_at.isoformat()
            }
        }, status=status.HTTP_201_CREATED)

            
    def accept_request(self, request, user_id=None):
        recipient = request.user
        print(recipient.id)
        try:
            follow_request = FollowRequest.objects.get(requester_id=user_id, recipient=recipient, status='pending')
            print(follow_request)
        except FollowRequest.DoesNotExist:
            return Response({"error": "Follow request not found or already handled"}, status=status.HTTP_404_NOT_FOUND)

        follow_request.status = 'accepted'
        follow_request.save()
        follow = Follow.objects.create(follower=follow_request.requester, following=recipient)

        # Notification.objects.filter(user=recipient, sender=follow_request.requester, action_type="follow_request").delete()
        Notification.objects.filter(user=recipient, sender=follow_request.requester, action_type="follow_request").delete()
        recipient_user = get_object_or_404(CustomUser, id=user_id)

        Notification.objects.create(
                user=recipient_user,
                sender=recipient,
                action_type="follow_accepted",
                message=f"{request.user.full_name} accepted your follow request."
            )
        
        privacy_settings, created = PrivacySettings.objects.get_or_create(user=recipient_user)
        # if recipient_user.privacy_settings.enable_notifications:
        if privacy_settings.enable_notifications:
            response = send_push_notification(
                fcm_token=recipient_user.fcm_token,
                title="Follow request accepted",
                body=f"{request.user.full_name} accepted your follow request."
            )
        else:
            response = None 


        sender_agora_username = request.user.agora_username
        receiver_agora_username = recipient_user.agora_username 

        app_token = ChatTokenBuilder.build_app_token(APP_ID, APP_CERTIFICATE, settings.AGORA_EXPIRE_TIME)
        agora_url_for_sender = f"{settings.AGORA_BASE_URL}/users/{sender_agora_username}/contacts/users/{receiver_agora_username}"
        agora_url_for_receiver = f"{settings.AGORA_BASE_URL}/users/{receiver_agora_username}/contacts/users/{sender_agora_username}"
        headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {app_token}",
        }
        response_for_sender = requests.post(agora_url_for_sender, headers=headers)
        if response_for_sender.status_code != 200:
                print("Agora API Error:", response_for_sender.status_code, response_for_sender.text)
                return Response({"error": "Failed to send message via Agora"}, status=status.HTTP_500_INTERNAL_SERVER_ERROR)
        response_for_receiver = requests.post(agora_url_for_receiver, headers=headers)
        if response_for_receiver.status_code != 200:
                print("Agora API Error:", response_for_receiver.status_code, response_for_receiver.text)
                return Response({"error": "Failed to send message via Agora"}, status=status.HTTP_500_INTERNAL_SERVER_ERROR)

        return Response({
            "message": "Follow request accepted",
            "push_notification": response,
            "data": {
                "user_id": str(follow.following.id),
                "follower_id": str(follow.follower.id),
                "followed_at": follow.followed_at.isoformat()
            }
        }, status=status.HTTP_200_OK)

    def cancel_request(self, request, user_id=None):
        """Allows the requester to cancel a pending follow request."""
        requester = request.user

        # Ensure only the requester can cancel their own follow request
        try:
            follow_request = FollowRequest.objects.get(
                requester=requester, recipient_id=user_id, status='pending'
            )
        except FollowRequest.DoesNotExist:
            return Response({"error": "Follow request not found or already handled"}, status=status.HTTP_404_NOT_FOUND)

        # Cancel the follow request if it's pending
        follow_request.status = 'cancelled'
        follow_request.save()
        Notification.objects.filter(user_id=user_id, sender=requester, action_type="follow_request").delete()
        recipient = CustomUser.objects.get(id=user_id)
        Notification.objects.create(
                user=recipient,
                sender=requester,
                action_type='other',
                message=f"Follow request expired {request.user.full_name}.",
                is_read=False
            )
        return Response({
            "success": True,
            "message": "Follow request canceled."
        }, status=status.HTTP_200_OK)

    def reject_request(self, request, user_id=None):
        """Allows the recipient to reject a pending follow request."""
        recipient = request.user

        # Ensure only the recipient can reject a follow request they received
        try:
            follow_request = FollowRequest.objects.get(
                requester_id=user_id, recipient=recipient, status='pending'
            )
        except FollowRequest.DoesNotExist:
            return Response({"error": "Follow request not found or already handled"}, status=status.HTTP_404_NOT_FOUND)

        # Reject the follow request if it's pending
        follow_request.status = 'rejected'
        follow_request.save()
        Notification.objects.filter(user=recipient, sender_id=user_id, action_type="follow_request").delete()
        return Response({
            "success": True,
            "message": "Follow request rejected."
        }, status=status.HTTP_200_OK)

    def unfollow(self, request, user_id=None):
        follower = request.user
        try:
            follow = Follow.objects.get(follower=follower, following_id=user_id)
            follow_status = FollowRequest.objects.get(recipient_id= user_id , requester_id =follower)
        except Follow.DoesNotExist:
            return Response({"error": "User is not currently following this user"}, status=status.HTTP_400_BAD_REQUEST)

        follow.delete()
        follow_status.delete()
        recipient_user = get_object_or_404(CustomUser, id=user_id)
        my_agora_username = follower.agora_username
        another_person_username = recipient_user.agora_username
        app_token = ChatTokenBuilder.build_app_token(APP_ID, APP_CERTIFICATE, settings.AGORA_EXPIRE_TIME)
        agora_url_for_sender = f"{settings.AGORA_BASE_URL}/users/{my_agora_username}/contacts/users/{another_person_username}"
        agora_url_for_receiver = f"{settings.AGORA_BASE_URL}/users/{another_person_username}/contacts/users/{my_agora_username}"
        headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {app_token}",
        }
        response_for_sender = requests.delete(agora_url_for_sender, headers=headers)
        if response_for_sender.status_code != 200:
                print("Agora API Error:", response_for_sender.status_code, response_for_sender.text)
                return Response({"error": "Failed to send message via Agora"}, status=status.HTTP_500_INTERNAL_SERVER_ERROR)
        response_for_receiver = requests.delete(agora_url_for_receiver, headers=headers)
        if response_for_receiver.status_code != 200:
                print("Agora API Error:", response_for_receiver.status_code, response_for_receiver.text)
                return Response({"error": "Failed to send message via Agora"}, status=status.HTTP_500_INTERNAL_SERVER_ERROR)

        return Response({
            "message": "Unfollowed successfully",
            "data": {
                "user_id": str(user_id),
                "follower_id": str(follower.id),
                "unfollowed_at": timezone.now().isoformat()
            }
        }, status=status.HTTP_200_OK)

    def list_requests(self, request):
        user = request.user
        follow_requests = FollowRequest.objects.filter(recipient=user, status='pending')
        data = []
        for req in follow_requests:
            requester = req.requester
            try:
                requester_profile = Profile.objects.get(user=requester)
                profile_pic_url = requester_profile.profile_image or ""
                role_name = requester_profile.selected_role.role_name if requester_profile.selected_role else ""
            except Profile.DoesNotExist:
                profile_pic_url = ""
                role_name = ""
            data.append({
                "requester_id": str(requester.id),
                "name": requester.username or "",
                "role": role_name,
                "profile_pic_url": profile_pic_url,
                "requested_at": req.created_at.isoformat()
            })
        return Response({
            "user_id": str(user.id),
            "follow_requests": data
        }, status=status.HTTP_200_OK)
    
def user_exists_in_agora(agora_username):
    app_token = ChatTokenBuilder.build_app_token(APP_ID, APP_CERTIFICATE, settings.AGORA_EXPIRE_TIME)
    agora_url = f"{settings.AGORA_BASE_URL}/users/{agora_username}"
    headers = {
        "Content-Type": "application/json",
        "Authorization": f"Bearer {app_token}",
    }
    response = requests.get(agora_url, headers=headers)
    return response.status_code == 200

class BlockViewSet(viewsets.ViewSet):
    permission_classes = [permissions.IsAuthenticated]

    def block_user(self, request, user_id=None):
        blocker = request.user
        if int(blocker.id) == int(user_id):
            return Response({"error": "You cannot block yourself."}, status=status.HTTP_400_BAD_REQUEST)

        try:
            blocked_user = User.objects.get(id=user_id)
        except User.DoesNotExist:
            return Response({"error": "User not found"}, status=status.HTTP_404_NOT_FOUND)

        if Block.objects.filter(blocker=blocker, blocked=blocked_user).exists():
            return Response({"error": "User is already blocked"}, status=status.HTTP_400_BAD_REQUEST)

        # Remove existing follow relationship
        Follow.objects.filter(follower=blocked_user, following=blocker).delete()
        Follow.objects.filter(follower=blocker, following=blocked_user).delete()

        # Create the block record
        common_reason = request.data.get("common_reason")
        reason_details = request.data.get("reason_details", "")
        block = Block.objects.create(
            blocker=blocker, 
            blocked=blocked_user, 
            common_reason=common_reason, 
            reason_details=reason_details
        )
        blocker_agora_username =  blocker.agora_username
        blocked_agora_username = blocked_user.agora_username

        if not user_exists_in_agora(blocker_agora_username):
            return Response({"error": "Blocker user not registered in Agora"}, status=status.HTTP_404_NOT_FOUND)
        if not user_exists_in_agora(blocked_agora_username):
            return Response({"error": "Blocked user not registered in Agora"}, status=status.HTTP_404_NOT_FOUND)
        blocked_user_list = []
        blocked_user_list.append(blocked_agora_username)
        print(blocker_agora_username,blocked_agora_username)

        app_token = ChatTokenBuilder.build_app_token(APP_ID, APP_CERTIFICATE, settings.AGORA_EXPIRE_TIME)
        agora_url = f"{settings.AGORA_BASE_URL}/users/{blocker_agora_username}/blocks/users"
        payload = {
            "usernames": blocked_user_list
        }
        headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {app_token}",
        }
        response_for_sender = requests.post(agora_url, json=payload, headers=headers)
        if response_for_sender.status_code != 200:
                print("Agora API Error:", response_for_sender.status_code, response_for_sender.text)
                return Response({"error": "Failed to Add in Block list via Agora"}, status=status.HTTP_500_INTERNAL_SERVER_ERROR)
        

        return Response({
            "success": True,
            "message": "User has been successfully blocked.",
            "data": {
                "blocked_user_id": str(blocked_user.id),
                "common_reason": common_reason,
                "reason_details": reason_details,
                "blocked_at": block.blocked_at.isoformat()
            }
        }, status=status.HTTP_201_CREATED)


    def unblock_user(self, request, user_id=None):
        blocker = request.user
        try:
            blocked_user = User.objects.get(id=user_id)
        except User.DoesNotExist:
            return Response({"error": "User not found"}, status=status.HTTP_404_NOT_FOUND)
        try:
            block = Block.objects.get(blocker=blocker, blocked_id=user_id)
        except Block.DoesNotExist:
            return Response({"error": "User not blocked"}, status=status.HTTP_404_NOT_FOUND)

        block.delete()
        # Agora API call to unblock the user
        blocker_agora_username = blocker.agora_username
        blocked_agora_username = blocked_user.agora_username

        app_token = ChatTokenBuilder.build_app_token(APP_ID, APP_CERTIFICATE, settings.AGORA_EXPIRE_TIME)
        agora_url = f"{settings.AGORA_BASE_URL}/users/{blocker_agora_username}/blocks/users/{blocked_agora_username}"
        headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {app_token}",
        }
        response = requests.delete(agora_url, headers=headers)
        if response.status_code != 200:
            print("Agora API Error:", response.status_code, response.text)
            return Response({"error": "Failed to remove from block list via Agora"}, status=status.HTTP_500_INTERNAL_SERVER_ERROR)

        return Response({
            "success": True,
            "message": "User unblocked successfully"
        }, status=status.HTTP_200_OK)

class ReportViewSet(viewsets.ViewSet):
    permission_classes = [permissions.IsAuthenticated]

    def report_user(self, request, user_id=None):
        reporter = request.user
        if int(reporter.id) == int(user_id):
            return Response({"error": "You cannot report yourself."}, status=status.HTTP_400_BAD_REQUEST)
        
        try:
            reported_user = User.objects.get(id=user_id)
        except User.DoesNotExist:
            return Response({"error": "User not found"}, status=status.HTTP_404_NOT_FOUND)

        data = request.data.copy()
        data['reported'] = user_id
        serializer = ReportSerializer(data=data)
        if serializer.is_valid():
            report = serializer.save(reporter=reporter)
            Notification.objects.create(
                user=reported_user,
                action_type='reported',
                message=f"Your account has been reported by {reporter.username}.",
                is_read=False
            )
            response = send_push_notification(
                fcm_token=reported_user.fcm_token,
                title="Account Reported",
                body=f"Your account has been reported by {reporter.username}."
            )
            return Response({
                "success": True,
                "push_notification": response,
                "message": "User has been successfully reported.",
                "data": serializer.data
            }, status=status.HTTP_201_CREATED)
        return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)


class BlockReasonListView(APIView):
    def get(self, request):
        reasons = [{'code': code, 'label': label} for code, label in Block.COMMON_BLOCK_REASONS]
        serializer = ReasonSerializer(reasons, many=True)
        return Response({"success": True, "data": serializer.data}, status=status.HTTP_200_OK)

class ReportReasonListView(APIView):
    def get(self, request):
        reasons = [{'code': code, 'label': label} for code, label in Report.COMMON_REPORT_REASONS]
        serializer = ReasonSerializer(reasons, many=True)
        return Response({"success": True, "data": serializer.data}, status=status.HTTP_200_OK)
    
class BlockedListView(APIView):
    permission_classes = [IsAuthenticated]

    def get(self, request):
        # Fetch the blocked list for the logged-in user
        blocked_users = Block.objects.filter(blocker=request.user)

        if blocked_users.exists():
            # Serialize the list of blocked users
            serializer = BlockListSerializer(blocked_users, many=True)
            return Response({
                "status": "success",
                "data": serializer.data
            }, status=status.HTTP_200_OK)
        else:
            # Return message if no data found
            return Response({
                "status": "success",
                "message": "No Data Found",
                "data": []
            }, status=status.HTTP_200_OK)
        
class FollowStatusView(APIView):
    permission_classes = [IsAuthenticated]

    def get(self, request):
        target_user_id = request.GET.get('user_id', None)

        if not target_user_id:
            return Response({
                "status": "failed",
                "message": "The 'user_id' query parameter is required."
            }, status=status.HTTP_400_BAD_REQUEST)

        try:
            print(request.user.id)
            is_following = Follow.objects.filter(follower=request.user, following_id=target_user_id).exists()
            follow_request = FollowRequest.objects.filter(requester=request.user, recipient_id=target_user_id).first()

            # Determine the status
            follow_status = follow_request.status if follow_request else None

            return Response({
                "status": "success",
                "is_following": is_following,
                "follow_request_status": follow_status  # Can be 'pending', 'accepted', 'cancelled', 'rejected', or None
            }, status=status.HTTP_200_OK)

        except Exception as e:
            return Response({
                "status": "failed",
                "message": f"An error occurred: {str(e)}"
            }, status=status.HTTP_500_INTERNAL_SERVER_ERROR)

class ReportStatusView(APIView):
    permission_classes = [IsAuthenticated]

    def get(self, request):
        target_user_id = request.GET.get('user_id', None)

        if not target_user_id:
            return Response({
                "status": "failed",
                "message": "The 'user_id' query parameter is required."
            }, status=status.HTTP_400_BAD_REQUEST)

        try:
            print(request.user.id)
            is_reported = Report.objects.filter(reporter_id=request.user, reported_id=target_user_id).exists()

            return Response({
                "status": "success",
                "is_reported": is_reported
            }, status=status.HTTP_200_OK)

        except Exception as e:
            return Response({
                "status": "failed",
                "message": f"An error occurred: {str(e)}"
            }, status=status.HTTP_500_INTERNAL_SERVER_ERROR)
        
class FollowerFollowingAPIView(APIView):
    permission_classes = [IsAuthenticated]

    def get(self, request, *args, **kwargs):
        user = request.user

        try:
            # Get the profile of the logged-in user
            profile = Profile.objects.get(user=user)
            follower_ids = Follow.objects.filter(following=user).values_list('follower__id', flat=True)
            following_ids = Follow.objects.filter(follower=user).values_list('following__id', flat=True)
            
            user_following_ids = set(following_ids)
            print(follower_ids)

            follower_profiles = Profile.objects.filter(
                user__id__in=follower_ids
            ).select_related('user').order_by('?')

            # Fetch profiles of following
            following_profiles = Profile.objects.filter(
                user__id__in=following_ids
            ).select_related('user').order_by('?')

            # Serialize the results
            followers = FollowerFollowingSerializer(
                follower_profiles, many=True, context={'request': request}
            ).data
            for follower in followers:
                follower["type"] = "follower"
                follower["is_following"] = follower["user_id"] in user_following_ids
                follow_request = FollowRequest.objects.filter(
                    requester=follower["user_id"], recipient_id=user
                ).first()
                follower["follow_status"] = follow_request.status if follow_request else None

            following = FollowerFollowingSerializer(
                following_profiles, many=True, context={'request': request}
            ).data
            for follow in following:
                follow["type"] = "following"
                follow["is_following"] = follow["user_id"] in user_following_ids
                follow_request = FollowRequest.objects.filter(
                    requester=user, recipient_id=follow["user_id"]
                ).first()
                print(user.id,follow["user_id"])
                follow["follow_status"] = follow_request.status if follow_request else None
            
            # Combine both lists
            combined = followers + following
            unique_users = {}
            for item in combined:
                if item["user_id"] not in unique_users:
                    unique_users[item["user_id"]] = item

            distinct_combined = list(unique_users.values())

            return Response({
                "status": "success",
                "data": distinct_combined
            }, status=200)

        except Profile.DoesNotExist:
            return Response({
                "error": "failed",
                "message": "Profile for the logged-in user does not exist."
            }, status=404)

        except Exception as e:
            return Response({
                "error": "failed",
                "message": f"An unexpected error occurred: {str(e)}"
            }, status=500)